/*
 * EcgLightC.c
 *
 * Created: 09.10.2012 15:36:17
 *  Author: Dima
 */ 

#include <avr/io.h>
#include <avr/interrupt.h>
#include "usart.c"
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include "commands.h"

#define uchar unsigned char


// device version and software version (WORDs)
// 6 - ecg light 2011
// 7 - lab controller
// 8 - ecg light 2013
#define DEVICE_VER1 8
// flash version
#define DEVICE_VER2 1

#define _MCU_CLOCK_FREQUENCY_  18432000
// MCU F in kHz
#define MCUCF 18432


typedef struct {
  uchar opcode; // as command + codes for ADC channels
  unsigned int data1;
  unsigned int data2;
} tReport;

typedef struct{
  uchar opcode;
  unsigned int data1;
  unsigned int data2;
} tCommand;

// for software calibration and serials
#define EEDataSize 32
uchar EEData[EEDataSize] EEMEM;

register uchar flags asm("r15");

unsigned char KETOvfCnt = 0;

#define FINDATAREADY  1
#define FOUTDATASENT  2
#define FOUTDATAREADY 4
#define FADCRUN       8

#define FlagSet(f) flags = flags | f
#define FlagClear(f) flags = flags & (~f)
#define IsFlag(f) (flags & f)

// port C
#define ADCRbit 0
#define ADCLbit 1

// port D
#define RXbit  0
#define TXbit  1
#define LEDbit 2


static inline void LedToggle(){
  if(PORTD & (1 << LEDbit)) 
    PORTD &= ~(1 << LEDbit); 
  else 
    PORTD |= (1 << LEDbit); 
}


unsigned char CurCh;
void ADC_SelectChannel(uchar ChN){
  ADMUX = (1<<REFS0)|(1<<REFS1)|(0<<ADLAR) | ChN;// ChN pin as input with internal 1.1 V ref
  CurCh = ChN;
}

void ADC_Init(void){ // needed for reference configuration
  //ADMUX = (1<<REFS0)|(1<<REFS1)|(0<<ADLAR)| ADCRbit;// ADC pin as input with internal 1.1 V ref
  ADC_SelectChannel(ADCRbit);
  ADCSRA = (1<<ADEN)|(1<<ADATE)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0); // autotrigger mode with clock / 128
  ADCSRB = 0; // free running mode
  ADCSRA |= (1<<ADIF)|(1<<ADSC); // start ADC
}


void TimersInit(){
  TCCR1A = 0;
  // Normal mode, clk/64 = 288 kHz
  TCCR1B = (1 << CS10) | (1 << CS11);
  // to get microseconds TCNT1 * 4 (for 16 MHz)
  // disable timer interrupts
  TIMSK1 = 0;  
}



void inline PutReport(tReport *rep){ 
  mputchars((uchar*)rep, sizeof(tReport));    
}


void RXdata(){       
   tCommand cmd;  
   tReport rep;
   while(rx_counter0 >= sizeof(cmd)){
      mgetchars((uchar*)&cmd, sizeof(cmd));
      // send standard report to the host about user command execution
      if(cmd.opcode < 99 || cmd.opcode > 105){
        rep.opcode = 100;
        rep.data1 = cmd.opcode;
        PutReport(&rep);  
      }  
      
      switch (cmd.opcode){                  
        // STANDARD COMMANDS
        case 99:        //stop                
          //disable tx of all ADC channels             
          FlagClear(FADCRUN);              
        break;
        
        case 101:       //ping            
            rep.opcode = 101;
            rep.data1 = cmd.data1;
            PutReport(&rep);  
            //LedToggle(); 
        break;
        
        case 102:       // version                  
            rep.opcode = 102;
            rep.data1 = DEVICE_VER1; 
            rep.data2 = DEVICE_VER2;
            PutReport(&rep);    
        break;  
        
        case 103:       // keepalive
            // if was not keepalive during some period - reset device state
            // do nothing here
        break;
        
        case 104: // eewrite - save 1 byte of user data in eeprom
            {
              uchar addr, data;
              addr = (cmd.data1 >> 8) & 0xFF;
              data = cmd.data1 & 0xFF;
              if(addr < EEDataSize){
                EEData[addr] = data;
              }
            }
        break;
        
        case 105: // eeread - read 1 byte of user data from eeprom
            {
              rep.opcode = 105;
              uchar addr, data;
              addr = cmd.data1; 
              rep.data1 = addr;
              if(addr < EEDataSize){
                data = EEData[addr];
                rep.data2 = data;
              }else{ // error
                rep.data2 = 0xFFFF; // error value
              }
              PutReport(&rep); 
            }
        break;
        
        
        // ADDITIONAL COMMANDS
        
        case ecom_adc_auto_run:       //start adc                   
            TCNT1 = 0;
            FlagSet(FADCRUN);                      
        break;       
        
        case ecom_adc_auto_stop:       //stop adc                   
            FlagClear(FADCRUN);                      
        break; 
        
        default: // ? error ?
          // cleanup buffers                    
          rx_wr_index0 = rx_rd_index0 = rx_counter0 = 0;                          
      } 
      KETOvfCnt = 0;            
   }
}

unsigned int ledi;
unsigned int LastWaitT;

void Capture(){  
  if(tx_counter0 > 0 || rx_counter0 >= sizeof(tCommand)) return; // there are queued data
  // wait to limit samplerate
  // timer F = 288 kHz. For samplerate 2 kHz wait 144 ticks; 3 kHz -> 96; 4 kHz -> 72
  unsigned int CurT, dt;
  do{
      CurT = TCNT1;
      if(CurT < LastWaitT) dt = 65535 - LastWaitT + CurT;
      else dt = CurT - LastWaitT;
      if(rx_counter0 > 0) return;
  }while(dt < 96);
  // apply correction and store last timestamp
  LastWaitT = CurT - (dt - 96);  
  
  // read and send a sample
  tReport rep;
  rep.data1 = CurT;
  rep.data2 = ADC;
  if(CurCh == ADCRbit){ // right is current   
    rep.opcode = ecom_adc_right; // Right Channel       
    ADC_SelectChannel(ADCLbit);// change channel
  }else{ // left
    rep.opcode = ecom_adc_left; // Left Channel    
    ADC_SelectChannel(ADCRbit);
  }  
  PutReport(&rep);  
  // led blinking in active mode
  ledi ++;
  if(ledi > 2000){
    LedToggle();      
    ledi = 0;
  }
}



int main(void)
{
    uchar   i = 0;  
   
    DDRC = 0; // all inputs
    PORTC = 0; // no pull-up
    
    DDRB = 0; // all inputs for unused port
    PORTB = 0;
    
    DDRD = (1 << TXbit) | (1 << LEDbit);
    PORTD = 0;
            
    wdt_disable();   
    cli();
    
    USARTInit();    
    TimersInit();      
    ADC_Init();    
          
    wdt_enable(WDTO_2S);   
    // start data handling
    sei();   
  
    LedToggle();    
    FlagClear(FADCRUN);
    
    unsigned int CurT;
    unsigned int KeepaliveLastT = 0;
    signed char ledinc = 1;
    unsigned int leddt, ledlastt = 0;
   
    while(1){  /* main event loop */                            
        i++;
        if(i == 0){
          wdt_reset();                         
        }  
        
        CurT = TCNT1;
        if(rx_counter0 >= sizeof(tCommand)){
          RXdata();          
        }else{         
          // check keepalive        
          if(KeepaliveLastT > CurT){            
            KETOvfCnt ++;
          }
          KeepaliveLastT = CurT;
                              
          if(KETOvfCnt > 30){ // keepalive timed out            
            KETOvfCnt = 0;        
            FlagClear(FADCRUN);
            rx_wr_index0 = rx_rd_index0 = rx_counter0 = 0;
            tx_wr_index0 = tx_rd_index0 = tx_counter0 = 0;              
          }
          
          if(IsFlag(FADCRUN)){
            Capture();            
            wdt_reset();
          }else{ // idle mode - make the led light on and off slowly
            if(CurT > ledlastt) leddt = CurT - ledlastt;
            else leddt = 65535 - ledlastt + CurT;
            if(leddt >= 1024){
              ledlastt = CurT;
              ledi += ledinc;
              if(ledi >= 1024){
                ledinc = -2;                                  
              }else if(ledi <= 2){
                ledinc = 2;
              } 
            }
            if(leddt < ledi) PORTD |= (1 << LEDbit);     
            else PORTD &= ~(1 << LEDbit);             
          }
        }        
    }    
}

/* ------------------------------------------------------------------------- */
